/*
    Copyright (C) 2015 William Hart

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
*/

#include "fmpz.h"
#include "gmp.h"
#include "fmpq.h"

void fmpq_get_mpz_frac(mpz_t a, mpz_t b, fmpq_t c)
{
   fmpz_get_mpz(a, fmpq_numref(c));
   fmpz_get_mpz(b, fmpq_denref(c));
}
