/*
    Copyright (C) 2007 David Howden
    Copyright (C) 2007, 2008, 2009, 2010 William Hart
    Copyright (C) 2008 Richard Howell-Peak
    Copyright (C) 2011 Fredrik Johansson
    Copyright (C) 2012 Lina Kulakova
    Copyright (C) 2013 Mike Hansen

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
*/

#ifdef T

#include "templates.h"

int
TEMPLATE(T, poly_is_irreducible) (const TEMPLATE(T, poly_t) f,
                                  const TEMPLATE(T, ctx_t) ctx)
{
    if (TEMPLATE(T, poly_length) (f, ctx) > 2)
    {
        return TEMPLATE(T, poly_is_irreducible_ddf) (f, ctx);
    }

    return 1;
}


#endif
