/*
    Copyright (C) 2017 Luca De Feo

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
*/

#include "fmpz_mod_mat.h"

void fmpz_mod_mat_trace(fmpz_t trace, const fmpz_mod_mat_t mat)
{
    fmpz_mat_trace(trace, mat->mat);
    fmpz_mod(trace, trace, mat->mod);
}
