/*
    Copyright (C) 2011 Sebastian Pancratz

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
*/

#include "padic.h"

void padic_ctx_clear(padic_ctx_t ctx)
{
    fmpz_clear(ctx->p);

    if (ctx->pow)
    {
        _fmpz_vec_clear(ctx->pow, ctx->max - ctx->min);
    }
}

